#
# Copyright (C) 2023-2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Command for creating PDF documents."""
import logging
from enum import Enum

from django.core.management.base import BaseCommand, CommandError, no_translations
from django.db import transaction
from django.utils import timezone
from filed import Storage
from typist import SecretaryClient

from django_pain.models import PaymentRequest
from django_pain.serializers import PaymentRequestContextSerializer
from django_pain.settings import get_file_storage_instance, get_secretary_client_instance

LOGGER = logging.getLogger(__name__)

PAYMENT_REQUEST_PDF_TEMPLATE_CZ = 'pain-payment-request-cs.html'
PAYMENT_REQUEST_PDF_TEMPLATE_EN = 'pain-payment-request-en.html'


class PdfType(Enum):
    """Types of PDF to create."""

    PDF_TYPE_PAYMENT_REQUEST = 'payment_request'


def render_payment_request_pdf(payment_request: PaymentRequest, secretary_client: SecretaryClient,
                               file_storage: Storage) -> None:
    """Render payment request to PDF and store the document with file manager."""
    if payment_request.file_uid is None:  # nothing to do if already rendered
        serializer = PaymentRequestContextSerializer(payment_request)
        if payment_request.country_code == 'CZ':
            pdf_template = PAYMENT_REQUEST_PDF_TEMPLATE_CZ
        else:
            pdf_template = PAYMENT_REQUEST_PDF_TEMPLATE_EN
        pdf = secretary_client.render_pdf(pdf_template, serializer.data)
        with file_storage.create(f'payment_request_{payment_request.uuid}.pdf', mimetype='application/pdf') as storage:
            storage.write(pdf)
        payment_request.file_uid = str(storage.uuid)
        payment_request.save(update_fields=['file_uid'])


class Command(BaseCommand):
    """Create PDF documents."""

    help = 'Create PDF documents.'

    def add_arguments(self, parser):
        """Add command arguments."""
        parser.add_argument('-t', '--type', choices=[i.value for i in PdfType], required=True, action='append',
                            help='Document type')

    @no_translations
    def handle(self, *args, **options):
        """Run command."""
        failed = 0
        LOGGER.info('Command `create_pdf` started')

        # Initialize clients to create and store PDF documents
        secretary_client = get_secretary_client_instance()
        file_storage = get_file_storage_instance()
        if secretary_client is None or file_storage is None:
            raise CommandError('Settings for secretary and file manager are required to create PDF documents')

        # Process payment requests
        if PdfType.PDF_TYPE_PAYMENT_REQUEST.value in options['type']:  # pragma: no branch  # temporary
            failed += self._handle_payment_requests(secretary_client, file_storage)

        LOGGER.info('Command `create_pdf` finished')
        if failed > 0:
            raise CommandError(f'Failed to create some PDFs: {failed}')

    @transaction.atomic
    def _handle_payment_requests(self, secretary_client: SecretaryClient, file_storage: Storage) -> int:
        """Create and store PDFs for payment requests."""
        LOGGER.info('Creating payment request PDFs ...')
        queryset = PaymentRequest.objects.filter(
            closed_at__isnull=True,  # active payment request
            due_at_displayed__gte=timezone.now(),  # not overdue
            file_uid__isnull=True,  # PDF not created yet
        ).select_for_update()
        failed = 0
        for payment_request in queryset:
            try:
                render_payment_request_pdf(payment_request, secretary_client, file_storage)
                LOGGER.debug('PDF for %s created', payment_request.uuid)
            except Exception as e:
                failed += 1
                LOGGER.error('PDF for %s not created: %r', payment_request.uuid, e)
        LOGGER.info('Payment request PDFs created: %d, failed: %d', len(queryset) - failed, failed)
        return failed
