#
# Copyright (C) 2018-2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Invoice related models."""
from typing import Optional, Tuple

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models, transaction
from django.db.models import CheckConstraint, Q, UniqueConstraint
from django.utils.translation import gettext_lazy as _
from django_countries.fields import CountryField
from djmoney.models.fields import CurrencyField

from django_pain.constants import CURRENCY_PRECISION, InvoiceType, PaymentProcessingError, PaymentType
from django_pain.settings import SETTINGS
from django_pain.utils import local_date

from .bank import BankAccount, BankPayment

INVOICE_TYPE_CHOICES = (
    (InvoiceType.ADVANCE, _('advance')),
    (InvoiceType.ACCOUNT, _('account')),
)


class InvoiceRef(models.Model):
    """Invoice reference model."""

    number = models.TextField(unique=True, verbose_name=_('Invoice number'))
    remote_id = models.IntegerField()
    invoice_type = models.TextField(choices=INVOICE_TYPE_CHOICES, verbose_name=_('Invoice type'))
    payments = models.ManyToManyField(BankPayment, related_name='invoice_refs')


# ----------------------------------------------------------------------------------------------------
# Customer model

class Customer(models.Model):
    """Customer model."""

    # Unique customer identifier
    uid = models.TextField(unique=True, verbose_name=_('Customer identifier'))
    # Typically, one of `name` or `organization` is set to easily distinguish physical/legal person
    name = models.TextField(null=True, blank=True, verbose_name=_('Name'))
    organization = models.TextField(null=True, blank=True, verbose_name=_('Organization'))
    # Customer address (to use on invoice)
    street1 = models.TextField(verbose_name=_('Street'))
    street2 = models.TextField(null=True, blank=True, verbose_name=_('Street 2'))
    street3 = models.TextField(null=True, blank=True, verbose_name=_('Street 3'))
    city = models.TextField(verbose_name=_('City'))
    state_or_province = models.TextField(null=True, blank=True, verbose_name=_('State or province'))
    postal_code = models.TextField(verbose_name=_('Postal code'))
    # Country represented by its two-letter ISO code
    country_code = CountryField(verbose_name=_('Country'))
    # Company registration number (like IČ in Czechia)
    company_registration_number = models.TextField(null=True, blank=True, verbose_name=_('Company registration number'))
    # VAT registration number (like DIČ in Czechia)
    vat_number = models.TextField(null=True, blank=True, verbose_name=_('VAT number'))
    # VAT was approved by VIES (`null` when not checked yet or not applicable)
    vies_check_ok = models.BooleanField(null=True, blank=True, verbose_name=_('VIES approved'))

    class Meta:
        """Meta class for customer model."""

        verbose_name = _('Customer')
        verbose_name_plural = _('Customers')

        constraints = [CheckConstraint(
            check=(
                ~(Q(name__isnull=True) | Q(name__exact=''))
                | ~(Q(organization__isnull=True) | Q(organization__exact=''))
            ),
            name='customer_name_or_organization_set'
        )]

    def __str__(self):
        """Return string representation of customer."""
        return f'{self.uid}: {self.name}' if self.name else f'{self.uid}: {self.organization}'


# ----------------------------------------------------------------------------------------------------
# Payment requst models

class PaymentRequestManager(models.Manager):
    """Customized manager for payment requests."""

    def find(self, app_id: str, payment: BankPayment
             ) -> Tuple[Optional['PaymentRequest'], Optional[PaymentProcessingError]]:
        """
        Find payment request related to given bank payment.

        Search is carried out on these criteria:
        - match in `app_id`
        - payment request is active, i.e. not closed
        - match in variable symbol
        - match in bank account for bank transfers / in identifier for card payments

        Return value consists of two parts:
        - payment request on successful search else `None`
        - optional error shows some discrepancy in payment amount (`INSUFFICIENT_AMOUNT` and `EXCESSIVE_AMOUNT`) or
        due time (`OVERDUE`)

        Note: Bank payment should be in `READY_TO_PROCESS` or `DEFERRED` state (ensured by caller).
        """
        payment_request = None
        error = None

        q = PaymentRequest.objects.filter(app_id=app_id, closed_at__isnull=True)
        if payment.payment_type == PaymentType.TRANSFER:
            q = q.filter(variable_symbol=payment.variable_symbol, bank_account=payment.account)
        else:  # PaymentType.CARD_PAYMENT
            q = q.filter(card_payment=payment)
        q.select_for_update()
        if q:  # related payment request found
            payment_request = q.get()

            due_date = local_date(payment_request.due_at) if payment_request.due_at is not None else None
            if due_date is not None and payment.transaction_date > due_date:
                error = PaymentProcessingError.OVERDUE
            elif payment.amount.amount < payment_request.price_to_pay:
                error = PaymentProcessingError.INSUFFICIENT_AMOUNT
            elif payment.amount.amount > payment_request.price_to_pay:
                error = PaymentProcessingError.EXCESSIVE_AMOUNT

        return payment_request, error


class PaymentRequest(models.Model):
    """
    Payment request model.

    Payment request provides two options to pay:
    - bank transfer: record doesn't appear as `BankPayment` until imported from bank statement
    - card payment: created on payment gateway with immediate `BankPayment` record
    Either one can be realized (in case of card payment `card_payment` and `realized_payment` fields
    refer to the same item).
    """

    uuid = models.UUIDField(unique=True)
    customer = models.ForeignKey(Customer, on_delete=models.RESTRICT, editable=False, related_name='+')
    app_id = models.TextField(verbose_name=_('Application ID'))
    # Copy of customer info at the moment of creation
    name = models.TextField(null=True, blank=True, verbose_name=_('Name'))
    organization = models.TextField(null=True, blank=True, verbose_name=_('Organization'))
    street1 = models.TextField(verbose_name=_('Street'))
    street2 = models.TextField(null=True, blank=True, verbose_name=_('Street 2'))
    street3 = models.TextField(null=True, blank=True, verbose_name=_('Street 3'))
    city = models.TextField(verbose_name=_('City'))
    state_or_province = models.TextField(null=True, blank=True, verbose_name=_('State or province'))
    postal_code = models.TextField(verbose_name=_('Postal code'))
    country_code = CountryField(verbose_name=_('Country'))
    company_registration_number = models.TextField(null=True, blank=True, verbose_name=_('Company registration number'))
    vat_number = models.TextField(null=True, blank=True, verbose_name=_('VAT number'))
    # Request lifetime related fields
    created_at = models.DateTimeField(auto_now_add=True, verbose_name=_('Create time'))
    due_at = models.DateTimeField(null=True, blank=True, verbose_name=_('Due time'))
    due_at_displayed = models.DateTimeField(verbose_name=_('Displayed due time'))
    closed_at = models.DateTimeField(null=True, blank=True, verbose_name=_('Close time'))
    # Total price to pay
    price_to_pay = models.DecimalField(max_digits=64, decimal_places=CURRENCY_PRECISION, verbose_name=_('Price to pay'))
    currency = CurrencyField()
    # Variable symbol
    variable_symbol = models.CharField(max_length=10, blank=True, verbose_name=_('Variable symbol'))
    # Bank account for bank transfer payment
    bank_account = models.ForeignKey(BankAccount, null=True, blank=True, on_delete=models.RESTRICT, related_name='+',
                                     db_column='ref_bank_account_id', verbose_name=_('Bank account'))
    # Bank payment (info about card payment when requested)
    card_payment = models.ForeignKey(BankPayment, null=True, blank=True, on_delete=models.RESTRICT, related_name='+',
                                     db_column='ref_card_payment_id', verbose_name=_('Card payment'))
    # Bank payment (info when payment was paid and successfully paired)
    realized_payment = models.ForeignKey(BankPayment, null=True, blank=True, on_delete=models.RESTRICT,
                                         related_name='+', verbose_name=_('Realized payment'))
    # PDF document in file manager storage
    file_uid = models.TextField(null=True, blank=True, verbose_name='PDF')

    objects = PaymentRequestManager()

    class Meta:
        """Model Meta class."""

        verbose_name = _('Payment request')
        verbose_name_plural = _('Payment requests')

        constraints = [
            # Variable symbol should be unique for all live (i.e. unclosed) payment requests
            UniqueConstraint(fields=['variable_symbol'], condition=Q(closed_at__isnull=True),
                             name='unique_vs_for_unclosed_payments')
        ]

    def __str__(self):
        """Return string representation."""
        return (f'{self.uuid}/{self.customer} {self.price_to_pay} {self.currency}'
                f' at {self.due_at_displayed.strftime("%Y-%m-%d")}')


class PaymentRequestItem(models.Model):
    """Model for payment request item."""

    # Parent payment_request
    payment_request = models.ForeignKey(PaymentRequest, on_delete=models.CASCADE, editable=False,
                                        related_name='items')
    # Descripton of the item
    name = models.TextField(verbose_name=_('Description'))
    # Item type (to evaluate VAT rate)
    item_type = models.TextField(null=True, blank=True, verbose_name=_('Item type'))
    # Quantity and pricing
    quantity = models.IntegerField(validators=(MinValueValidator(1),), default=1, verbose_name=_('Quantity'))
    price_per_unit = models.DecimalField(max_digits=64, decimal_places=CURRENCY_PRECISION,
                                         verbose_name=_('Price per unit'))
    price_total = models.DecimalField(max_digits=64, decimal_places=CURRENCY_PRECISION, verbose_name=_('Price total'))

    def __str__(self):
        """Return string representation."""
        return f'{self.payment_request.uuid}/{self.name}'


class BillingItemType(models.Model):
    """Type of item for billing, VAT rate evaluation, etc."""

    handle = models.TextField(verbose_name=_('Name'))
    # Validity range
    valid_from = models.DateTimeField(verbose_name=_('Time from'))
    valid_to = models.DateTimeField(null=True, blank=True, verbose_name=_('Time to'))  # `null` for unlimited
    # VAT rate in per cents
    vat_rate = models.DecimalField(max_digits=10, decimal_places=4, verbose_name=_('VAT rate'))

    def __str__(self):
        """Return string representation."""
        return f'{self.handle} - {self.vat_rate}%'


class VariableSymbolSequenceManager(models.Manager):
    """Customized manager for variable symbol sequences."""

    @transaction.atomic
    def next_value(self, app_id: str, year: int) -> str:
        """Get variable symbol from appropriate sequence."""
        # Prevent variable symbol duplicates by locking this into DB transaction
        vs_sequence: VariableSymbolSequence = self.filter(app_id=app_id, year=year).select_for_update().get()
        vs_sequence.current_number += 1
        vs_sequence.save(update_fields=['current_number'])  # ToDo: Handle sequence exhaustion
        return vs_sequence.variable_symbol


class VariableSymbolSequence(models.Model):
    """Variable symbol sequence model."""

    app_id = models.TextField(verbose_name=_('Application ID'))
    year = models.IntegerField(validators=(MinValueValidator(1), MaxValueValidator(9_999)), verbose_name=_('Year'))
    prefix = models.IntegerField(validators=(MinValueValidator(0), MaxValueValidator(9_999)),
                                 verbose_name=_('Prefix'))
    current_number = models.IntegerField(editable=False, default=0)

    objects = VariableSymbolSequenceManager()

    class Meta:
        """Meta class for model."""

        unique_together = ['app_id', 'year', 'prefix']

        constraints = [
            CheckConstraint(check=Q(year__gte=1) & Q(year__lte=9_999), name='vs_year_range'),
            CheckConstraint(check=Q(prefix__gte=0) & Q(prefix__lte=9_999), name='vs_prefix_4_digits'),
            CheckConstraint(check=Q(current_number__gte=0) & Q(current_number__lte=999_999),
                            name='vs_current_number_6_digits'),
        ]

    @property
    def variable_symbol(self) -> str:
        """Format variable symbol as string (up to 10 digits)."""
        variable_symbol = f'{self.prefix:0>4d}{self.current_number:0>6d}'
        if SETTINGS.trim_varsym:
            variable_symbol = variable_symbol.lstrip('0')
        return variable_symbol

    def __str__(self):
        """Return string representation of VS sequence."""
        return f'{self.app_id}: {self.variable_symbol}'
