#
# Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Serializers for REST API."""
import logging
import re
from datetime import datetime, timedelta
from decimal import Decimal
from enum import Enum, unique
from re import match as regex_match
from typing import Any, Dict, List, Optional, OrderedDict, Union

from django.conf import settings
from django.db.models import Q
from django.utils import timezone
from django_countries import countries
from djmoney.money import Money
from rest_framework import serializers

from django_pain.card_payment_handlers import CartItem, PaymentHandlerLimitError
from django_pain.constants import PaymentState
from django_pain.models import BankAccount, BankPayment, Customer, PaymentRequest, PaymentRequestItem
from django_pain.models.invoices import BillingItemType, VariableSymbolSequence
from django_pain.settings import SETTINGS, get_card_payment_handler_instance
from django_pain.utils import local_date, round_money_amount

LOGGER = logging.getLogger(__name__)


@unique
class ExternalPaymentState(str, Enum):
    """Card Payment state used by REST API."""

    INITIALIZED = 'initialized'
    PAID = 'paid'
    CANCELED = 'canceled'


CARD_PAYMENT_STATE_MAPPING = {
    PaymentState.INITIALIZED: ExternalPaymentState.INITIALIZED,
    PaymentState.READY_TO_PROCESS: ExternalPaymentState.PAID,
    PaymentState.PROCESSED: ExternalPaymentState.PAID,
    PaymentState.DEFERRED: ExternalPaymentState.PAID,
    PaymentState.EXPORTED: ExternalPaymentState.PAID,
    PaymentState.CANCELED: ExternalPaymentState.CANCELED,
    PaymentState.TO_REFUND: ExternalPaymentState.PAID,
    PaymentState.REFUNDED: ExternalPaymentState.PAID,
}


def custom_expiry_validator(expiry: datetime) -> None:
    """Validate custom expiry."""
    now = timezone.now()
    min_expiry = timedelta(seconds=SETTINGS.csob_custom_handler['min_validity'])
    max_expiry = timedelta(seconds=SETTINGS.csob_custom_handler['max_validity'])
    if not (now + min_expiry <= expiry <= now + max_expiry):
        raise serializers.ValidationError('Time point outside allowed interval')


class BankPaymentSerializer(serializers.ModelSerializer):
    """Serializer for BankPayment."""

    return_url = serializers.URLField(write_only=True)
    return_method = serializers.ChoiceField(choices=('GET', 'POST'), write_only=True)
    cart = serializers.JSONField(write_only=True, required=True)
    language = serializers.CharField(write_only=True)
    identifier = serializers.CharField(read_only=True)
    state = serializers.SerializerMethodField(read_only=True)
    gateway_redirect_url = serializers.SerializerMethodField(read_only=True)
    email = serializers.EmailField(required=False, write_only=True)
    custom_expiry = serializers.DateTimeField(required=False, write_only=True, validators=(custom_expiry_validator,))

    class Meta:
        model = BankPayment
        fields = ['uuid', 'identifier', 'amount', 'variable_symbol', 'processor', 'card_handler',
                  'return_url', 'return_method', 'cart', 'language', 'gateway_redirect_url', 'state',
                  'email', 'custom_expiry']

    def __init__(self, *args, **kwargs):
        self.gateway_redirect_url = None
        super().__init__(*args, **kwargs)

    def get_state(self, obj):
        """Get simple state for REST API from BankPayment state attribute."""
        return CARD_PAYMENT_STATE_MAPPING[obj.state]

    def get_gateway_redirect_url(self, obj):
        """Get redriect URL of the payment gateway."""
        return self.gateway_redirect_url

    def validate_cart(self, value):
        """Validate cart is properly set."""
        if not (1 <= len(value) <= 2):
            raise serializers.ValidationError('`cart` must have one or two item(s)!')
        cart_items = []
        for item in value:
            if len(item['name']) > 20:
                raise serializers.ValidationError('`cartitem/name` must not exceed 20 characters!')
            cart_item = CartItem(**item)
            cart_items.append(cart_item)
        return cart_items

    def validate(self, data):
        """Overall checks throughout the fields."""
        data = self._validate_amount(data)
        data = self._validate_amount_cart(data)
        return data

    def _validate_amount(self, data):
        """
        Validate amount is of type Money.

        If amount_currency is not specified we get Decimal instead of Money.
        We can not validate amount in the dedicated function because we want to log info about the processor.
        """
        value = data['amount']
        if not isinstance(value, Money):
            processor = data.get('processor', '')
            message = 'Parameter "amount_currency" not set. Using default currency. Processor for this payment: "{}"'
            LOGGER.warning(message.format(processor))
            data['amount'] = Money(amount=value, currency=settings.DEFAULT_CURRENCY)
        return data

    def _validate_amount_cart(self, data):
        """Validate cart and amount are consistent."""
        cart_total_amount = sum((item.amount * item.quantity for item in data['cart']))
        if cart_total_amount != data['amount'].amount:
            raise serializers.ValidationError('Sum of `cart` items amounts must be equal to payments `amount`')
        return data

    def create(self, validated_data):
        """Create the payment using initialization of its CardHandler."""
        card_handler = get_card_payment_handler_instance(validated_data.pop('card_handler'))
        payment, self.gateway_redirect_url = card_handler.init_payment(**validated_data)
        return payment


class StreetField(serializers.Field):
    """
    Serializer for `street` field.

    Street is stored as three separate items in data model (`street1`, `street2` and `street3`)
    while passed as a list of (non-empty) strings in API. At least one non-empty string is required.
    """

    def street_validator(self, value: Any) -> None:
        """Validate list of one to three non-empty strings."""
        if not isinstance(value, list) or len(value) < 1 or len(value) > 3:
            raise serializers.ValidationError('List of one to three strings is required')
        for i in value:
            if not isinstance(i, str) or not i:
                raise serializers.ValidationError('List items should be non-empty strings')

    def to_representation(self, value: Union[Customer, PaymentRequest]) -> List[str]:
        """Convert to list of strings."""
        result = [value.street1]  # `street1` is non-empty
        for i in (value.street2, value.street3):
            if i:
                result.append(i)
            else:
                break
        return result

    def to_internal_value(self, data: List[str]) -> Dict[str, Optional[str]]:
        """Validate and convert from list of strings."""
        self.street_validator(data)
        result = {
            'street1': data[0],
            'street2': data[1] if len(data) > 1 else None,
            'street3': data[2] if len(data) > 2 else None,
        }
        return result


class CustomerSerializer(serializers.ModelSerializer):
    """Serializer for `Customer` model."""

    street = StreetField(source='*')

    class Meta:
        model = Customer
        fields = ['uid', 'name', 'organization', 'street', 'city', 'state_or_province', 'postal_code', 'country_code',
                  'company_registration_number', 'vat_number']

    def validate(self, data: OrderedDict) -> OrderedDict:
        """Perform object-wide validations."""
        # In case of partial update `data` may not be enough to perform object-wide validation properly and
        # should be combined with stored model instance data.
        if self.partial and self.instance:
            full_data = self.to_representation(self.instance)
            full_data.update(data)
        else:
            full_data = data
        if not full_data.get('name') and not full_data.get('organization'):
            raise serializers.ValidationError('Neither `name` nor `organization` is set')
        return data


class PaymentRequestItemSerializer(serializers.ModelSerializer):
    """Serializer for `PaymentRequstItem` model."""

    price_with_vat = serializers.BooleanField(write_only=True)

    class Meta:
        model = PaymentRequestItem
        fields = ['name', 'item_type', 'quantity', 'price_per_unit', 'price_total', 'price_with_vat']
        read_only_fields = ['price_total']

    def validate(self, data: OrderedDict) -> OrderedDict:
        """Perform object-wide validations and processing."""
        price_with_vat = data.pop('price_with_vat')
        data['price_per_unit'] = round_money_amount(data['price_per_unit'])

        # Validate item type and get VAT rate from code list
        now = timezone.now()
        q = BillingItemType.objects.filter(handle=data['item_type'], valid_from__lte=now).filter(
            Q(valid_to__gt=now) | Q(valid_to__isnull=True)
        )
        if not q:
            raise serializers.ValidationError('No applicable item type and VAT rate found')
        if len(q) > 1:
            raise serializers.ValidationError('Multiple applicable VAT rates found')

        # Add VAT to price if required
        if not price_with_vat:
            data['price_per_unit'] = round_money_amount(
                data['price_per_unit'] * (Decimal(1) + Decimal('0.01') * q.get().vat_rate)
            )

        # Calculate total price
        data['price_total'] = data['quantity'] * data['price_per_unit']

        return data


def not_in_past_validator(timestamp: Optional[datetime]) -> None:
    """Validate that timestamp in not in the past."""
    if timestamp is not None and timestamp < timezone.now():
        raise serializers.ValidationError('Time point must not be in the past')


def vs_format_validator(variable_symbol: str) -> None:
    """Validate format of variable symbol."""
    # Check format
    pattern = r'^[0-9]{0,10}$'
    if regex_match(pattern, variable_symbol) is None:
        raise serializers.ValidationError('Invalid format')


def variable_symbol_validator(variable_symbol: str) -> None:
    """Validate variable symbol."""
    # Ensure there is no active payment request with the same symbol
    q = PaymentRequest.objects.filter(closed_at=None, variable_symbol=variable_symbol).select_for_update()
    if q:
        raise serializers.ValidationError(
            f'Duplicate variable symbol found in active payment request `{q.get().uuid}`'
        )
    # Ensure there is a closed payment request with the same symbol
    # to prevent possible conflict with the symbol sequence.
    q = PaymentRequest.objects.filter(closed_at__isnull=False, variable_symbol=variable_symbol)
    if not q:
        raise serializers.ValidationError(
            f'Closed payment request with variable symbol `{variable_symbol}` not found')


class CardPaymentSerializer(serializers.Serializer):
    """Serializer for card payment parameters."""

    expiry = serializers.DateTimeField(allow_null=True, default=None, validators=(custom_expiry_validator,))
    redirect_url = serializers.URLField()
    redirect_method = serializers.ChoiceField(choices=['GET', 'POST'], default='GET')


class PaymentRequestSerializer(serializers.ModelSerializer):
    """Serializer for `PaymentRequst` model."""

    customer = serializers.SlugRelatedField(slug_field='uid', queryset=Customer.objects.all())
    street = StreetField(source='*', read_only=True)
    bank_account = serializers.SlugRelatedField(slug_field='account_number', queryset=BankAccount.objects.all())
    variable_symbol = serializers.CharField(
        max_length=10, allow_null=True, default=None, validators=[vs_format_validator, variable_symbol_validator]
    )  # None => generate VS
    due_at = serializers.DateTimeField(allow_null=True, required=False, validators=(not_in_past_validator,))
    due_at_displayed = serializers.DateTimeField(validators=(not_in_past_validator,))
    card_payment = serializers.CharField(source='card_payment.uuid', read_only=True, allow_null=True)
    realized_payment = serializers.CharField(source='realized_payment.uuid', read_only=True, allow_null=True)
    card_gateway_url = serializers.SerializerMethodField(read_only=True)

    items = PaymentRequestItemSerializer(many=True, min_length=1)

    create_card_payment = CardPaymentSerializer(default=None, allow_null=True, write_only=True)

    class Meta:
        model = PaymentRequest
        fields = ['uuid', 'app_id', 'customer', 'name', 'organization', 'street', 'city', 'state_or_province',
                  'postal_code', 'country_code', 'company_registration_number', 'vat_number', 'created_at', 'due_at',
                  'due_at_displayed', 'closed_at', 'price_to_pay', 'currency', 'variable_symbol', 'bank_account',
                  'card_payment', 'realized_payment', 'card_gateway_url', 'items', 'create_card_payment', 'file_uid']
        read_only_fields = ['name', 'organization', 'street', 'city', 'state_or_province', 'postal_code',
                            'country_code', 'company_registration_number', 'vat_number', 'created_at', 'closed_at',
                            'price_to_pay', 'file_uid']

    def get_card_gateway_url(self, obj: PaymentRequest) -> Optional[str]:
        """Get URL on the bank payment gateway to pay the request."""
        result = None
        if obj.card_payment:
            handler = get_card_payment_handler_instance(obj.card_payment.card_handler)
            result = handler.get_payment_url(obj.card_payment)

        return result

    def validate(self, data: OrderedDict) -> OrderedDict:
        """Perform object-wide validations and processing."""
        # Validate account currency compatibility
        if data.get('bank_account') and (data['currency'] != data['bank_account'].currency):
            raise serializers.ValidationError('Selected currency is not compatible with the bank account')

        # Copy current customer data
        customer = data['customer']
        data['name'] = customer.name
        data['organization'] = customer.organization
        data['street1'] = customer.street1
        data['street2'] = customer.street2
        data['street3'] = customer.street3
        data['city'] = customer.city
        data['state_or_province'] = customer.state_or_province
        data['postal_code'] = customer.postal_code
        data['country_code'] = customer.country_code
        data['company_registration_number'] = customer.company_registration_number
        data['vat_number'] = customer.vat_number

        # Calculate total price as sum of all items
        data['price_to_pay'] = sum(i['price_total'] for i in data['items'])

        return data

    def create(self, validated_data: OrderedDict[str, Any]) -> PaymentRequest:
        """Create payment request object with all its items."""
        items = validated_data.pop('items')
        create_card_payment = validated_data.pop('create_card_payment')

        # Variable symbol processing
        if validated_data['variable_symbol'] is None:
            # Variable symbol is to be generated
            app_id = validated_data['app_id']
            year = timezone.now().year
            try:
                validated_data['variable_symbol'] = VariableSymbolSequence.objects.next_value(app_id, year)
            except VariableSymbolSequence.DoesNotExist:
                raise serializers.ValidationError(
                    {'detail': f'No variable symbol sequence available for `{app_id}` and `{year}`'},
                )

        # Save objects to database
        payment_request = PaymentRequest.objects.create(**validated_data)
        for item in items:
            PaymentRequestItem.objects.create(payment_request=payment_request, **item)

        # Create card payment on bank gateway if required
        if create_card_payment is not None:
            settings = SETTINGS.payment_requests[validated_data['app_id']]
            handler = get_card_payment_handler_instance(settings['CARD_HANDLER'])
            params = {
                'amount': Money(validated_data['price_to_pay'], validated_data['currency']),
                'variable_symbol': validated_data['variable_symbol'],
                'processor': settings['PROCESSOR'],
                'return_url': create_card_payment['redirect_url'],
                'return_method': create_card_payment['redirect_method'],
                'cart': self._make_cart(items),
                'language': 'cs',
            }
            if create_card_payment['expiry']:
                params['custom_expiry'] = create_card_payment['expiry']

            # Payment request is created without card payment when amount exceeds limit given by bank and
            # no error is reported in such case.
            try:
                payment_request.card_payment, _ = handler.init_payment(**params)
                payment_request.save(update_fields=['card_payment'])
            except PaymentHandlerLimitError:
                pass

        return payment_request

    def _make_cart(self, items: List[OrderedDict[str, Any]]) -> List[CartItem]:
        """Convert payment request items to cart for card payment."""
        # Cart must have one or two values, description up to 20 chars
        item = items[0]  # `items` is sure to be non-empty
        result = [CartItem(
            name=item['name'][:20], quantity=item['quantity'],
            amount=float(item['price_per_unit'] * item['quantity']), description=''),
        ]
        items_left = len(items) - 1
        if items_left == 1:
            item = items[1]
            result.append(CartItem(name=item['name'][:20], quantity=item['quantity'],
                                   amount=float(item['price_per_unit'] * item['quantity']), description=''))
        elif items_left > 1:  # accumulate left items into one auxuliary cart item
            result.append(CartItem(
                name=f'+{items_left} ...',
                quantity=1,
                amount=float(sum(i['price_total'] for i in items[1:])),
                description='',
            ))
        return result


class PaymentRequestContextSerializer(serializers.ModelSerializer):
    """Read only serializer for `PaymentRequst` model to create context for rendering to document."""

    street = StreetField(source='*')
    bank_account = serializers.SlugRelatedField(slug_field='account_number', queryset=BankAccount.objects.all())

    country = serializers.SerializerMethodField()
    created_at = serializers.SerializerMethodField()
    due_at = serializers.SerializerMethodField()
    price_paid = serializers.SerializerMethodField()
    iban = serializers.SerializerMethodField()
    qr_code = serializers.SerializerMethodField()
    items = serializers.SerializerMethodField()

    # Czech bank account number: optional prefix number + prefix separator, base number, slash, bank code
    cz_account_pattern = re.compile(r'((?P<prefix>\d{2,6})\s*-\s*)?(?P<base>\d{2,10})\s*\/\s*(?P<bank>\d{4})')

    class Meta:
        model = PaymentRequest
        fields = ['name', 'organization', 'street', 'city', 'state_or_province', 'postal_code', 'country',
                  'company_registration_number', 'vat_number', 'variable_symbol', 'created_at', 'due_at',
                  'bank_account', 'price_paid', 'price_to_pay', 'currency', 'iban', 'qr_code', 'items']
        read_only_fields = fields

    def get_created_at(self, obj: PaymentRequest) -> str:
        """Format created time as local date."""
        return local_date(obj.created_at).isoformat()

    def get_due_at(self, obj: PaymentRequest) -> str:
        """Format due time as local date."""
        return local_date(obj.due_at_displayed).isoformat()

    def get_country(self, obj: PaymentRequest) -> str:
        """Return country name."""
        return dict(countries)[obj.country_code]

    def get_price_paid(self, obj: PaymentRequest) -> str:
        """Pass amount already paid."""
        return '0.00'  # Because of workflow no amount is paid in advance.

    def _account_to_iban(self, account: str) -> str:
        """Convert account number to IBAN format."""
        # Beware: national format to IBAN conversion is Czechia specific!
        match = self.cz_account_pattern.match(account)
        if match is None:
            raise ValueError('Invalid bank account number: {}'.format(account))
        # IBAN format for Czechia at https://www.cnb.cz/cs/platebni-styk/iban/iban-mezinarodni-format-cisla-uctu/
        iban_account = '{bank}{prefix:0>6}{base:0>10}'.format(
            prefix=match.group('prefix') or 0,
            base=match.group('base'),
            bank=match.group('bank'),
        )
        # Compute checksum
        check_string = iban_account + '123500'  # from 'CZ00', C => 12 Z => 35
        check_number = 98 - int(check_string) % 97
        # Compose result
        return 'CZ{:0>2d}{}'.format(check_number, iban_account)

    def get_iban(self, obj: PaymentRequest) -> str:
        """Return bank account in IBAN format."""
        result = self._account_to_iban(obj.bank_account.account_number)
        return ' '.join([result[i:i + 4] for i in range(0, len(result), 4)])  # form with space separated groups

    def get_qr_code(self, obj: PaymentRequest) -> str:
        """Return SPAYD string to generate QR code for bank transfer."""
        # Format specification at https://cbaonline.cz/upload/1645-standard-qr-v1-2-cerven-2021.pdf
        data = (
            'SPD',  # header
            '1.0',  # format version
            'ACC:{}'.format(self._account_to_iban(obj.bank_account.account_number)),  # bank account
            'AM:{}'.format(round_money_amount(obj.price_to_pay)),  # amount (always 2 decimal places)
            'CC:{}'.format(obj.currency),  # currency
            'DT:{}'.format(obj.due_at_displayed.strftime('%Y%m%d')),  # due date
            'X-VS:{}'.format(obj.variable_symbol),  # variable symbol
            'MSG:CZ.NIC'  # message
        )
        return '*'.join(data)

    def get_items(self, obj: PaymentRequest) -> List[Dict[str, Any]]:
        """Return payment request items."""
        items = PaymentRequestItem.objects.filter(payment_request=obj.pk)
        return [
            {
                'name': i.name,
                'quantity': i.quantity,
                'price_total': str(i.price_total),
            } for i in items
        ]
