#
# Copyright (C) 2018-2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Test models."""
from datetime import date, datetime
from decimal import Decimal
from typing import Any, Dict, List, Tuple

from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.db.models import BLANK_CHOICE_DASH
from django.test import SimpleTestCase, TestCase, override_settings
from djmoney.money import Money
from freezegun import freeze_time

from django_pain.constants import InvoiceType, PaymentProcessingError, PaymentState, PaymentType
from django_pain.models import (BankPayment, PaymentImportHistory, PaymentRequest, ProcessPaymentResultType,
                                VariableSymbolSequence)
from django_pain.models.invoices import BillingItemType, PaymentRequestItem
from django_pain.utils import ProcessPaymentResult

from .mixins import CacheResetMixin
from .utils import create_payment_request, get_account, get_customer, get_invoice_ref, get_payment, get_payment_request


class TestBankAccount(SimpleTestCase):
    """Test BankAccount model."""

    def test_str(self):
        """Test string representation."""
        account = get_account(account_name='Account', account_number='123')
        self.assertEqual(str(account), 'Account 123')


class TestBankPayment(CacheResetMixin, TestCase):
    """Test BankPayment model."""

    def test_str(self):
        """Test string representation."""
        payment = get_payment(identifier='PID')
        self.assertEqual(str(payment), 'PID')

    def test_currency_constraint_success(self):
        """Test clean method with not violated constraint."""
        account = get_account(currency='USD')
        payment = get_payment(account=account, amount=Money('999.00', 'USD'))
        payment.clean()

    def test_currency_constraint_error(self):
        """Test clean method with violated constraint."""
        account = get_account(account_name='ACCOUNT', account_number='123', currency='USD')
        payment = get_payment(identifier='PAYMENT', account=account, amount=Money('999.00', 'CZK'))
        with self.assertRaisesMessage(ValidationError, 'Bank payment PAYMENT is in different currency (CZK) '
                                                       'than bank account ACCOUNT 123 (USD).'):
            payment.clean()

    @override_settings(PAIN_PROCESSORS={'dummy': 'django_pain.tests.utils.DummyPaymentProcessor'})
    def test_objective_choices(self):
        self.assertEqual(BankPayment.objective_choices(), BLANK_CHOICE_DASH + [
            ('dummy', 'Dummy objective'),
        ])

    def test_advance_invoice(self):
        account = get_account()
        account.save()
        payment = get_payment(account=account)
        payment.save()

        account_invoice = get_invoice_ref(number='1', invoice_type=InvoiceType.ACCOUNT)
        account_invoice.save()
        account_invoice.payments.add(payment)
        self.assertEqual(payment.advance_invoice, None)

        advance_invoice = get_invoice_ref(number='2', invoice_type=InvoiceType.ADVANCE)
        advance_invoice.save()
        advance_invoice.payments.add(payment)
        self.assertEqual(payment.advance_invoice, advance_invoice)

    @override_settings(PAIN_PROCESSORS={'dummy': 'django_pain.tests.utils.DummyPaymentProcessor'})
    def test_objective(self):
        payment = get_payment()
        self.assertEqual(payment.objective, '')

        payment = get_payment(processor='dummy')
        self.assertEqual(payment.objective, 'Dummy objective')

    def test_transfer_may_not_have_counter_account(self):
        account = get_account()
        account.save()

        payment = get_payment(account=account, payment_type=PaymentType.TRANSFER, counter_account_number='')
        payment.save()

        self.assertEqual(BankPayment.objects.get().counter_account_number, '')

    def test_card_payment_must_not_have_counter_account(self):
        account = get_account()
        account.save()

        payment = get_payment(account=account, payment_type=PaymentType.CARD_PAYMENT, counter_account_number='123')
        self.assertRaises(IntegrityError, payment.save)

    def test_custom_payment(self):
        account = get_account()
        account.save()
        payment = get_payment(
            account=account,
            payment_type=PaymentType.CARD_PAYMENT,
            counter_account_number='',
            email='user@example.org',
            custom_expiry=datetime(2023, 7, 31, 12, 0, 0),
            customer_code='ABC123',
            email_sent=datetime(2023, 8, 1, 9, 0, 0),
        )
        payment.save()

        payment = BankPayment.objects.get()
        self.assertEqual(payment.email, 'user@example.org')
        self.assertEqual(payment.custom_expiry, datetime(2023, 7, 31, 12, 0, 0))
        self.assertEqual(payment.customer_code, 'ABC123')
        self.assertEqual(payment.email_sent, datetime(2023, 8, 1, 9, 0, 0))

    def test_custom_payment_defaults(self):
        account = get_account()
        account.save()
        payment = get_payment(
            account=account,
            payment_type=PaymentType.CARD_PAYMENT,
            counter_account_number='',
        )
        payment.save()

        payment = BankPayment.objects.get()
        self.assertIsNone(payment.email)
        self.assertIsNone(payment.custom_expiry)
        self.assertEqual(payment.customer_code, '')
        self.assertIsNone(payment.email_sent)

    def test_save_processing_result(self):
        data = [
            ('id-1', ProcessPaymentResultType.PROCESSED, None, PaymentState.PROCESSED),
            ('id-2', ProcessPaymentResultType.TO_REFUND, None, PaymentState.TO_REFUND),
            ('id-3', ProcessPaymentResultType.NOT_PROCESSED, PaymentProcessingError.OVERDUE, PaymentState.DEFERRED),
        ]
        account = get_account()
        account.save()
        for id, result, error, state in data:
            with self.subTest(id=id):
                payment = get_payment(identifier=id, account=account)

                payment.save_processing_result(ProcessPaymentResult(result=result, error=error))

                payment = BankPayment.objects.get(identifier=id)
                self.assertEqual(payment.state, state)
                self.assertEqual(payment.processing_error, error)


class TestPaymentImportHistory(CacheResetMixin, TestCase):
    """Test PaymentImportHistory model."""

    @freeze_time('2021-02-01 10:15')
    def test_str(self):
        """Test string representation."""
        import_history = PaymentImportHistory(origin='test')
        import_history.save()
        self.assertEqual(str(import_history), 'test 2021-02-01 10:15:00')

    @freeze_time('2021-02-01 10:15')
    def test_can_not_override_start_datetime(self):
        """Test string representation."""
        import_history = PaymentImportHistory(origin='test', start_datetime=datetime(2021, 4, 1))
        import_history.save()
        self.assertEqual(str(import_history), 'test 2021-02-01 10:15:00')

    def test_filenames(self):
        import_history = PaymentImportHistory(origin='test')
        self.assertEqual(list(import_history.filenames), [])

        import_history.add_filename('file_1.txt')
        self.assertEqual(list(import_history.filenames), ['file_1.txt'])

        import_history.add_filename('file_2.txt')
        self.assertEqual(list(import_history.filenames), ['file_1.txt', 'file_2.txt'])

    def test_filename_can_not_contain_separator(self):
        import_history = PaymentImportHistory(origin='test')
        with self.assertRaisesRegex(ValueError, 'Character ; not allowed in filenames'):
            import_history.add_filename('file;name.txt')

    def test_success(self):
        """Test success property."""
        self.assertTrue(PaymentImportHistory(origin='test', errors=0, finished=True).success)
        self.assertFalse(PaymentImportHistory(origin='test', errors=1, finished=True).success)
        self.assertFalse(PaymentImportHistory(origin='test', errors=None, finished=True).success)
        self.assertFalse(PaymentImportHistory(origin='test', errors=0, finished=False).success)


class TestCustomer(SimpleTestCase):
    """Test `Customer` model."""

    def test_str(self):
        """Test string representation."""
        data = [
            ('John Doe', None, 'CID-1: John Doe'),
            ('John Doe', 'The Company, Ltd.', 'CID-1: John Doe'),
            (None, 'The Company, Ltd.', 'CID-1: The Company, Ltd.'),
            ('', 'The Company, Ltd.', 'CID-1: The Company, Ltd.'),
        ]
        for name, organization, expected in data:
            with self.subTest(name=name, organization=organization):
                customer = get_customer(name=name, organization=organization)
                self.assertEqual(str(customer), expected)


class TestVariableSymbolSequence(SimpleTestCase):
    """Test model for variable symbol sequence."""

    def test_str(self):
        """Test string representation."""
        data: List[Tuple[Dict[str, Any], str]] = [
            ({'prefix': 0}, '0000000000'),
            ({'prefix': 1, 'current_number': 12345}, '0001012345'),
            ({'prefix': 1234}, '1234000000'),
        ]
        for num_items, expected in data:
            with self.subTest(num_items=num_items):
                items = {'app_id': 'test'}
                items.update(num_items)
                vs_sequence = VariableSymbolSequence(**items)
                self.assertEqual(str(vs_sequence), f'test: {expected}')

    @override_settings(PAIN_TRIM_VARSYM=True)
    def test_str_trimmed(self):
        """Test string representation."""
        data: List[Tuple[Dict[str, Any], str]] = [
            ({'prefix': 0}, ''),
            ({'prefix': 1, 'current_number': 12345}, '1012345'),
            ({'prefix': 1234}, '1234000000'),
        ]
        for num_items, expected in data:
            with self.subTest(num_items=num_items):
                items = {'app_id': 'test'}
                items.update(num_items)
                vs_sequence = VariableSymbolSequence(**items)
                self.assertEqual(str(vs_sequence), f'test: {expected}')


class TestVariableSymbolSequenceManager(TestCase):
    """Test of retrieving variable_symbol from sequence."""

    def setUp(self) -> None:
        vs_item = VariableSymbolSequence(app_id='test', year=2023, prefix=823, current_number=205)
        vs_item.save()

    def test_get_ok(self):
        self.assertEqual(VariableSymbolSequence.objects.next_value('test', 2023), '0823000206')
        vs_sequence = VariableSymbolSequence.objects.get(app_id='test', year=2023)
        self.assertEqual(vs_sequence.current_number, 206)

    def test_sequence_not_found(self):
        with self.assertRaises(VariableSymbolSequence.DoesNotExist):
            VariableSymbolSequence.objects.next_value('other_test', 2023)

    @override_settings(PAIN_TRIM_VARSYM=True)
    def test_get_ok_trimmed(self):
        self.assertEqual(VariableSymbolSequence.objects.next_value('test', 2023), '823000206')


class TestPaymentRequest(SimpleTestCase):
    """Test `PaymentRequest` model."""

    def test_str(self):
        """Test string representation."""
        self.assertEqual(
            str(get_payment_request()),
            '00000000-0000-0000-0000-000000000001/CID-1: Test, s.r.o. 25.60 CZK at 2023-10-01',
        )


class TestPaymentRequestManager(TestCase):
    """Test customized manager for `PaymentRequest` model."""

    def test_found_ok(self):
        payment_request = create_payment_request(realized_payment=None)
        payment = get_payment(
            identifier='pid-1',
            account=payment_request.bank_account,
            variable_symbol=payment_request.variable_symbol,
            amount=Money(payment_request.price_to_pay, payment_request.currency),
            transaction_date=date(2023, 10, 5),
        )
        payment.save()
        payment = BankPayment.objects.get(identifier='pid-1')

        found, error = PaymentRequest.objects.find('auctions', payment)
        self.assertEqual(payment_request, found)
        self.assertIsNone(error)

    def test_not_found_closed(self):
        payment_request = create_payment_request(realized_payment=None, closed_at=datetime(2023, 9, 15))
        payment = get_payment(
            identifier='pid-1',
            account=payment_request.bank_account,
            variable_symbol=payment_request.variable_symbol,
            amount=Money(payment_request.price_to_pay, payment_request.currency),
            transaction_date=date(2023, 10, 5),
        )
        payment.save()

        found, _ = PaymentRequest.objects.find('auctions', payment)
        self.assertIsNone(found)

    def test_not_found_vs(self):
        payment_request = create_payment_request(realized_payment=None)
        payment = get_payment(
            identifier='pid-1',
            account=payment_request.bank_account,
            variable_symbol='1234567890',
            amount=Money(payment_request.price_to_pay, payment_request.currency),
            transaction_date=date(2023, 10, 5),
        )
        payment.save()

        found, _ = PaymentRequest.objects.find('auctions', payment)
        self.assertIsNone(found)

    def test_found_overdue(self):
        payment_request = create_payment_request(realized_payment=None, due_at=datetime(2023, 10, 1))
        payment = get_payment(
            identifier='pid-1',
            account=payment_request.bank_account,
            variable_symbol=payment_request.variable_symbol,
            amount=Money(payment_request.price_to_pay, payment_request.currency),
            transaction_date=date(2023, 10, 5),
        )
        payment.save()

        found, error = PaymentRequest.objects.find('auctions', payment)
        self.assertEqual(payment_request, found)
        self.assertIs(error, PaymentProcessingError.OVERDUE)

    def test_found_insufficient_amount(self):
        payment_request = create_payment_request(realized_payment=None)
        payment = get_payment(
            identifier='pid-1',
            account=payment_request.bank_account,
            variable_symbol=payment_request.variable_symbol,
            amount=Money(payment_request.price_to_pay - 1, payment_request.currency),
            transaction_date=date(2023, 10, 5),
        )
        payment.save()

        found, error = PaymentRequest.objects.find('auctions', payment)
        self.assertEqual(payment_request, found)
        self.assertIs(error, PaymentProcessingError.INSUFFICIENT_AMOUNT)

    def test_found_excessive_amount(self):
        payment_request = create_payment_request(realized_payment=None)
        payment = get_payment(
            identifier='pid-1',
            account=payment_request.bank_account,
            variable_symbol=payment_request.variable_symbol,
            amount=Money(payment_request.price_to_pay + 1, payment_request.currency),
            transaction_date=date(2023, 10, 5),
        )
        payment.save()

        found, error = PaymentRequest.objects.find('auctions', payment)
        self.assertEqual(payment_request, found)
        self.assertIs(error, PaymentProcessingError.EXCESSIVE_AMOUNT)

    def test_found_card_payment(self):
        payment_request = create_payment_request(realized_payment=None)
        payment = get_payment(
            identifier='pid-1',
            payment_type=PaymentType.CARD_PAYMENT,
            account=payment_request.bank_account,
            counter_account_number='',
            counter_account_name='',
            variable_symbol=payment_request.variable_symbol,
            amount=Money(payment_request.price_to_pay, payment_request.currency),
            transaction_date=date(2023, 10, 5),
        )
        payment.save()
        payment_request.card_payment = payment
        payment_request.save(update_fields=['card_payment'])

        found, error = PaymentRequest.objects.find('auctions', payment)
        self.assertEqual(payment_request, found)
        self.assertIsNone(error)


class TestPaymentRequestItem(SimpleTestCase):
    """Test `PaymentRequestItem` model."""

    def test_str(self):
        """Test string representation."""
        payment_request_item = PaymentRequestItem(
            payment_request=get_payment_request(),
            name='Test item',
            item_type='basic_rate',
            quantity=2,
            price_per_unit=Decimal('8.05'),
            price_total=Decimal('16.10'),
        )
        self.assertEqual(str(payment_request_item), '00000000-0000-0000-0000-000000000001/Test item')


class TestBillingItemType(SimpleTestCase):
    """Test `BillingItemType` model."""

    def test_str(self):
        """Test string representation."""
        billing_item_type = BillingItemType(
            handle='basic_rate',
            valid_from=datetime(2020, 1, 1),
            valid_to=None,
            vat_rate=Decimal('18.0'),
        )
        self.assertEqual(str(billing_item_type), 'basic_rate - 18.0%')
