#
# Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Tests of serializers."""
from collections import OrderedDict
from datetime import datetime
from decimal import Decimal
from typing import Any, Dict, List
from unittest.mock import patch
from uuid import UUID

from django.test import SimpleTestCase, TestCase, override_settings
from djmoney.money import Money
from freezegun import freeze_time
from rest_framework.serializers import ErrorDetail, ValidationError
from testfixtures import LogCapture

from django_pain.card_payment_handlers import AbstractCardPaymentHandler, CartItem, PaymentHandlerLimitError
from django_pain.constants import PaymentState, PaymentType
from django_pain.models import BankAccount, BankPayment, PaymentRequest
from django_pain.models.invoices import BillingItemType, VariableSymbolSequence
from django_pain.serializers import (BankPaymentSerializer, CustomerSerializer, ExternalPaymentState,
                                     PaymentRequestContextSerializer, PaymentRequestItemSerializer,
                                     PaymentRequestSerializer, custom_expiry_validator, not_in_past_validator,
                                     vs_format_validator)
from django_pain.tests.utils import (create_payment_request, get_account, get_customer, get_payment,
                                     get_payment_request, get_payment_request_context_data, get_payment_request_data)


class TestBankPaymentSerializer(SimpleTestCase):
    """
    Test BankPayment serializer.
    """

    def setUp(self):
        self.log_handler = LogCapture(('django_pain.serializers',), propagate=False)

    def tearDown(self):
        self.log_handler.uninstall()

    def test_get_state(self):
        account = get_account(account_number='123456', currency='CZK')
        payment = get_payment(identifier='1', account=account, counter_account_name='Account one',
                              state=PaymentState.PROCESSED)

        serializer = BankPaymentSerializer()
        self.assertEqual(serializer.get_state(payment), ExternalPaymentState.PAID)

    def test_validate_cart(self):
        serializer = BankPaymentSerializer()

        self.assertRaisesRegex(ValidationError, 'one or two item', serializer.validate_cart, [])
        self.assertRaisesRegex(ValidationError, 'one or two item', serializer.validate_cart, [{}, {}, {}])

        self.assertRaisesRegex(ValidationError, 'must not exceed 20', serializer.validate_cart,
                               [{'name': 'Dar too long to fit max length',
                                 'amount': 20, 'description': 'Dar datovce', 'quantity': 1}])

        result = serializer.validate_cart([{'name': 'Dar',
                                            'amount': 20,
                                            'description': 'Dar datovce',
                                            'quantity': 1}])
        self.assertEqual(result, [CartItem(name='Dar', amount=20, description='Dar datovce', quantity=1)])

    @override_settings(DEFAULT_CURRENCY='EUR')
    def test_validate_amount(self):
        data: Dict[str, Any] = {
            'cart': [CartItem(name='Dar', amount=10, description='Dar datovce', quantity=1)]
        }
        serializer = BankPaymentSerializer()

        data['amount'] = Money(10, 'CZK')
        validated = serializer.validate(data)
        self.assertEqual(validated['amount'], Money(10, 'CZK'))

        data['amount'] = Decimal(10)
        validated = serializer.validate(data)
        self.assertEqual(validated['amount'], Money(10, 'EUR'))

        data['amount'] = 10
        data['processor'] = 'processor'
        validated = serializer.validate(data)
        self.assertEqual(validated['amount'], Money(10, 'EUR'))

        message = 'Parameter "amount_currency" not set. Using default currency. Processor for this payment: "{}"'
        self.log_handler.check(
            ('django_pain.serializers', 'WARNING', message.format('')),
            ('django_pain.serializers', 'WARNING', message.format('processor')),
        )

    def test_validate_amount_and_cart(self):
        data = {
            'amount': Money(40, 'CZK'),
            'cart': [CartItem(name='Dar', amount=20, description='Dar datovce', quantity=2)]
        }
        serializer = BankPaymentSerializer()
        self.assertEqual(data, serializer.validate(data))

        data['amount'] = Money(100, 'CZK')
        self.assertRaisesRegex(ValidationError, 'amounts must be equal', serializer.validate, data)


class TestStreetFieldSerializer(TestCase):
    """`StreetField` related tests."""

    def setUp(self) -> None:
        self.items = {
            'uid': 'CID-1',
            'street': ['Na Okraji 1'],
            'city': 'Horní Dolní',
            'postal_code': '050 10',
            'country_code': 'CZ',
            'company_registration_number': '001234',
            'vat_number': 'CZ001234',
        }

    def test_street_validation_ok(self):
        data: List[Any] = [
            ['Street 1'],
            ['Street 1', 'Street 2'],
            ['Street 1', 'Street 2', 'Street 3'],
        ]
        for street in data:
            with self.subTest(street=street):
                self.items.update({'name': 'Jan Novák', 'street': street})
                serializer = CustomerSerializer(data=self.items)
                self.assertTrue(serializer.is_valid())

    def test_street_validation_bad_list(self):
        data: List[Any] = [
            123,  # not a list
            [],  # list too short
            ['Street 1', 'Street 2', 'Street 3', 'Street 4'],  # list too long
        ]
        for street in data:
            with self.subTest(street=street):
                self.items.update({'name': 'Jan Novák', 'street': street})
                serializer = CustomerSerializer(data=self.items)
                self.assertFalse(serializer.is_valid())
                self.assertIn(
                    ErrorDetail(string='List of one to three strings is required', code='invalid'),
                    serializer.errors['street'],
                )

    def test_street_validation_bad_list_items(self):
        data: List[Any] = [
            ['Street 1', 'Street 2', ''],  # empty string item
            ['Street 1', True],  # non-string item
        ]
        for street in data:
            with self.subTest(street=street):
                self.items.update({'name': 'Jan Novák', 'street': street})
                serializer = CustomerSerializer(data=self.items)
                self.assertFalse(serializer.is_valid())
                self.assertIn(
                    ErrorDetail(string='List items should be non-empty strings', code='invalid'),
                    serializer.errors['street'],
                )

    def test_street_serialization(self):
        data = [
            ('Street 1', 'Street 2', 'Street 3', ['Street 1', 'Street 2', 'Street 3']),
            ('Street 1', 'Street 2', None, ['Street 1', 'Street 2']),
            ('Street 1', None, None, ['Street 1']),
        ]
        for street1, street2, street3, expected in data:
            with self.subTest(street1=street1, street2=street2, street3=street3):
                customer = get_customer(street1=street1, street2=street2, street3=street3)
                serializer = CustomerSerializer(customer)
                self.assertEqual(serializer.data['street'], expected)


class TestCustomerSerializer(TestCase):
    """`CustomerSerializer` tests."""

    def setUp(self) -> None:
        self.items = {
            'uid': 'CID-1',
            'street': ['Na Okraji 1'],
            'city': 'Horní Dolní',
            'postal_code': '050 10',
            'country_code': 'CZ',
            'company_registration_number': '001234',
            'vat_number': 'CZ001234',
        }

    def test_validation_ok(self):
        data: List[Dict[str, Any]] = [
            {'name': 'Jan Novák'},
            {'organization': 'Firma, s.r.o.'},
            {'name': 'Jan Novák', 'organization': 'Firma, s.r.o.'},
        ]
        for new_items in data:
            with self.subTest(new_item=new_items):
                items = self.items
                items.update(new_items)
                serializer = CustomerSerializer(data=items)
                self.assertTrue(serializer.is_valid())

    def test_validation_neither_name_nor_organization(self):
        data: List[Dict[str, Any]] = [
            {},
            {'name': ''},
            {'name': None},
            {'organization': ''},
            {'organization': None},
            {'name': '', 'organization': ''},
            {'name': None, 'organization': None},
        ]
        for new_items in data:
            with self.subTest(new_items=new_items):
                items = self.items
                items.update(new_items)
                serializer = CustomerSerializer(data=items)
                self.assertFalse(serializer.is_valid())
                self.assertIn(
                    ErrorDetail(string='Neither `name` nor `organization` is set', code='invalid'),
                    serializer.errors['non_field_errors'],
                )


class TestPaymentRequestItemSerializer(TestCase):
    """Tests for validation, VAT processing and price calculation in `PaymentRequestItemSerializer`."""

    def setUp(self) -> None:
        # Initialize code list
        BillingItemType.objects.create(
            handle='test_type', valid_from=datetime(2023, 1, 1), valid_to=datetime(2023, 5, 1), vat_rate=Decimal('13.5')
        )
        BillingItemType.objects.create(
            handle='test_type', valid_from=datetime(2023, 1, 1), valid_to=datetime(2023, 2, 1), vat_rate=Decimal('13.5')
        )
        BillingItemType.objects.create(
            handle='test_type', valid_from=datetime(2023, 5, 1), valid_to=None, vat_rate=Decimal('18')
        )
        # Create related objects
        customer = get_customer()
        customer.save()
        account = get_account()
        account.save()
        payment_request = get_payment_request(customer=customer, bank_account=account)
        payment_request.save()

        self.data = {'payment_request': payment_request, 'name': 'Item 1', 'item_type': 'test_type', 'quantity': 2,
                     'price_per_unit': '1.0000000000', 'price_with_vat': False}

    def test_validation_ok(self):
        test_data = [
            ('2023-03-01', Decimal('1.14'), Decimal('2.28')),  # VAT rate 13.5%
            ('2023-08-01', Decimal('1.18'), Decimal('2.36')),  # VAT rate 18%
        ]
        for when, expected_unit_price, expected_total_price in test_data:
            with self.subTest(when=when):
                with freeze_time(when):
                    serializer = PaymentRequestItemSerializer(data=self.data)
                    self.assertTrue(serializer.is_valid())
                    self.assertEqual(serializer.validated_data['price_per_unit'], expected_unit_price)
                    self.assertEqual(serializer.validated_data['price_total'], expected_total_price)

    def test_validation_failure_time(self):
        test_data = [
            ('2022-10-01', 'No applicable item type and VAT rate found'),
            ('2023-01-15', 'Multiple applicable VAT rates found'),
        ]
        for when, expected_message in test_data:
            with self.subTest(when=when):
                with freeze_time(when):
                    serializer = PaymentRequestItemSerializer(data=self.data)
                    self.assertFalse(serializer.is_valid())
                    self.assertIn(
                        ErrorDetail(string=expected_message, code='invalid'), serializer.errors['non_field_errors'],
                    )

    @freeze_time('2023-06-01T12:00:00')
    def test_validation_failure_item_type(self):
        data = self.data
        data.update(item_type='another_type', price_with_vat=True)
        serializer = PaymentRequestItemSerializer(data=data)
        self.assertFalse(serializer.is_valid())
        self.assertIn(
            ErrorDetail(string='No applicable item type and VAT rate found', code='invalid'),
            serializer.errors['non_field_errors'],
        )


class TestNotInPastValidator(SimpleTestCase):
    """Tests of timestamp validation (not to be in past)."""

    def test_validation_none(self):
        not_in_past_validator(None)

    @freeze_time('2024-02-05T17:00:00')
    def test_validation_ok(self):
        not_in_past_validator(datetime(2024, 2, 15))

    @freeze_time('2024-02-05T17:00:00')
    def test_validation_raises(self):
        with self.assertRaisesMessage(ValidationError, 'Time point must not be in the past'):
            not_in_past_validator(datetime(2024, 1, 15))


class TestVariableSmbolFormatValidator(TestCase):
    """Tests of variable symbol format validation."""

    def test_ok(self):
        data = [
            '',  # empty
            '1234',  # short
            '0123456789',  # full
        ]
        for vs in data:
            with self.subTest(vs=vs):
                vs_format_validator(vs)

    def test_invalid(self):
        data = [
            '0123456789012',  # too long
            ' 1234',  # invalid character
        ]
        for vs in data:
            with self.subTest(vs=vs):
                self.assertRaises(ValidationError, vs_format_validator, vs)


@override_settings(PAIN_CSOB_CUSTOM_HANDLER={'MIN_VALIDITY': 3600, 'MAX_VALIDITY': 86400})
@freeze_time("2023-01-01 00:00:00")
class TestCustomExpiryValidator(SimpleTestCase):
    """Tests of validator for custom expiry."""

    def test_custom_expiry_ok(self):
        custom_expiry_validator(datetime(2023, 1, 1, 10, 0, 0))

    def test_custom_expiry_invalid(self):
        data = [
            datetime(2023, 1, 1, 0, 59, 59),  # too early
            datetime(2023, 1, 2, 0, 0, 1),  # too late
        ]

        for custom_expiry in data:
            with self.subTest(custom_expiry=custom_expiry):
                with self.assertRaisesMessage(ValidationError, 'Time point outside allowed interval'):
                    custom_expiry_validator(custom_expiry)


class TestCardPaymentHandler(AbstractCardPaymentHandler):
    """Card payment handler for tests of payment requests."""

    def init_payment(self, amount, variable_symbol, processor, return_url, return_method, cart, language, **kwargs):
        """Create and save test payment (with some data based on variable symbol)."""
        payment = BankPayment.objects.create(
            identifier='pid-' + variable_symbol,
            uuid=UUID(int=int(variable_symbol)),
            payment_type=PaymentType.CARD_PAYMENT,
            account=BankAccount.objects.get(account_number='123456/0300'),
            amount=amount,
            description='Test payment',
            state=PaymentState.INITIALIZED,
            card_payment_state='',
            variable_symbol=variable_symbol,
            processor=processor,
            card_handler=self.name,
            **kwargs,
        )
        return payment, 'https://gateway.example.org/' + variable_symbol

    def update_payments_state(self, payment):
        """Do nothing."""

    def get_payment_url(self, payment):
        """Return URL."""
        return 'https://gateway.example.org/' + payment.variable_symbol


class TestPaymentRequestSerializer(TestCase):
    """`PaymentRequestSerializer` tests."""

    # --- Serialization tests ---

    def test_serialization(self):
        # Prepare data and save all objects to database
        pk = create_payment_request().pk

        # Get object from database and serialize it
        serializer = PaymentRequestSerializer(PaymentRequest.objects.get(pk=pk))
        self.assertEqual(serializer.data, get_payment_request_data())

    def _prepare_get_card_gateway_url(self):
        account = get_account(account_number='1000/1000')
        account.save()
        data = {
            'identifier': 'PAYMENT1',
            'payment_type': PaymentType.CARD_PAYMENT,
            'account': account,
            'amount': Money('42.00', 'CZK'),
            'description': 'Gift for FRED',
            'state': PaymentState.INITIALIZED,
            'variable_symbol': '2023000001',
            'card_handler': 'custom',
            'email': 'user@example.org',
            'custom_expiry': datetime(2023, 10, 15, 15, 30, 0),
            'customer_code': 'XYZ789',
        }
        card_payment = BankPayment(**data)
        card_payment.save()
        return create_payment_request(card_payment=card_payment)

    @override_settings(PAIN_CSOB_CUSTOM_HANDLER={}, PAIN_CARD_PAYMENT_HANDLERS={
        'custom': {'HANDLER': 'django_pain.card_payment_handlers.csob.CSOBCustomPaymentHandler', 'GATEWAY': 'default'},
    })
    @freeze_time('2023-09-20')
    def test_get_card_gateway_url(self):
        payment_request = self._prepare_get_card_gateway_url()
        serializer = PaymentRequestSerializer(payment_request)
        self.assertEqual(serializer.data['card_gateway_url'], 'https://platebnibrana.csob.cz/zaplat/XYZ789')

    @override_settings(PAIN_CSOB_CUSTOM_HANDLER={}, PAIN_CARD_PAYMENT_HANDLERS={
        'dummy': {'HANDLER': 'django_pain.tests.utils.DummyCardPaymentHandler', 'GATEWAY': 'dummy'},
    })
    def test_get_card_gateway_url_default(self):
        payment_request = self._prepare_get_card_gateway_url()
        payment_request.card_payment.card_handler = 'dummy'
        serializer = PaymentRequestSerializer(payment_request)
        self.assertIsNone(serializer.data['card_gateway_url'])

    # --- Deserialization and validation tests ---

    @freeze_time('2023-09-01T12:00:00')
    def test_create_ok(self):
        create_payment_request(
            uuid='00000000-0000-0000-0000-000000000002',
            closed_at='2023-09-20T12:00:00',  # stored payment request is closed
        )
        BillingItemType.objects.create(
            handle='basic_rate', valid_from=datetime(2023, 1, 1), valid_to=None, vat_rate=Decimal('13.5')
        )

        data: Dict[str, Any] = {
            'uuid': '00000000-0000-0000-0000-000000000001',
            'app_id': 'auctions',
            'customer': 'CID-1',
            'due_at_displayed': '2023-10-01T12:00:00',
            'currency': 'CZK',
            'variable_symbol': '0000009876',
            'bank_account': '123456/0300',
            'items': [
                {'name': 'Item 1', 'item_type': 'basic_rate', 'quantity': 1, 'price_per_unit': '11.0000000000',
                 'price_with_vat': True},
                {'name': 'Item 2', 'item_type': 'basic_rate', 'quantity': 2, 'price_per_unit': '12.0000000000',
                 'price_with_vat': True},
                {'name': 'Item 3', 'item_type': 'basic_rate', 'quantity': 3, 'price_per_unit': '13.0000000000',
                 'price_with_vat': True},
            ],
        }

        # Create and save objects to database
        serializer = PaymentRequestSerializer(data=data)
        self.assertTrue(serializer.is_valid())
        serializer.save()

        # Serialize stored objects for verificaton
        serializer = PaymentRequestSerializer(PaymentRequest.objects.get(uuid=UUID(int=1)))
        self.assertEqual(serializer.data, get_payment_request_data(realized_payment=None))

    @freeze_time('2023-09-01T12:00:00')
    def test_create_ok_with_created_vs(self):
        customer = get_customer()
        customer.save()
        account = get_account()
        account.save()
        vs_item = VariableSymbolSequence(app_id='auctions', year=2023, prefix=823, current_number=205)
        vs_item.save()
        BillingItemType.objects.create(
            handle='basic_rate', valid_from=datetime(2023, 1, 1), valid_to=None, vat_rate=Decimal('13.5')
        )

        data: Dict[str, Any] = {
            'uuid': '00000000-0000-0000-0000-000000000001',
            'app_id': 'auctions',
            'customer': 'CID-1',
            'due_at_displayed': '2023-10-01T12:00:00',
            'currency': 'CZK',
            'bank_account': '123456/0300',
            'items': [
                {'name': 'Item 1', 'item_type': 'basic_rate', 'quantity': 1, 'price_per_unit': '11.0000000000',
                 'price_with_vat': True},
            ],
        }

        serializer = PaymentRequestSerializer(data=data)
        self.assertTrue(serializer.is_valid())
        self.assertIsNone(serializer.validated_data['variable_symbol'])
        payment_request = serializer.save()  # VS generated just before saving to DB
        self.assertEqual(payment_request.variable_symbol, '0823000206')

    @freeze_time('2023-09-01T12:00:00')
    def test_bad_field(self):
        customer = get_customer()
        customer.save()
        account = get_account()
        account.save()

        data: Dict[str, Any] = {
            'uuid': '00000000-0000-0000-0000-000000000001',
            'app_id': None,
            'customer': 'CID-1',
            'due_at_displayed': '2023-10-01T12:00:00',
            'currency': 'CZK',
            'variable_symbol': '0000009876',
            'bank_account': '123456/0300',
            'items': [
                {'name': 'Item 1', 'item_type': 'basic_rate', 'quantity': 1, 'price_per_unit': '11.0000000000',
                 'price_with_vat': True},
            ],
        }

        serializer = PaymentRequestSerializer(data=data)
        self.assertFalse(serializer.is_valid())
        self.assertIn(
            ErrorDetail(string='This field may not be null.', code='null'),
            serializer.errors['app_id'],
        )

    @freeze_time('2023-09-01T12:00:00')
    def test_customer_not_found(self):
        account = get_account()
        account.save()

        data: Dict[str, Any] = {
            'uuid': '00000000-0000-0000-0000-000000000001',
            'app_id': 'auctions',
            'customer': 'CID-1',
            'due_at_displayed': '2023-10-01T12:00:00',
            'currency': 'CZK',
            'variable_symbol': '0000009876',
            'bank_account': '123456/0300',
            'items': [
                {'name': 'Item 1', 'item_type': 'basic_rate', 'quantity': 1, 'price_per_unit': '11.0000000000',
                 'price_with_vat': True},
            ],
        }

        serializer = PaymentRequestSerializer(data=data)
        self.assertFalse(serializer.is_valid())
        self.assertIn(
            ErrorDetail(string='Object with uid=CID-1 does not exist.', code='does_not_exist'),
            serializer.errors['customer'],
        )

    @freeze_time('2023-09-01T12:00:00')
    def test_account_not_found(self):
        customer = get_customer()
        customer.save()

        data: Dict[str, Any] = {
            'uuid': '00000000-0000-0000-0000-000000000001',
            'app_id': 'auctions',
            'customer': 'CID-1',
            'due_at_displayed': '2023-10-01T12:00:00',
            'currency': 'CZK',
            'variable_symbol': '0000009876',
            'bank_account': '123456/0300',
            'items': [
                {'name': 'Item 1', 'item_type': 'basic_rate', 'quantity': 1, 'price_per_unit': '11.0000000000',
                 'price_with_vat': True},
            ],
        }

        serializer = PaymentRequestSerializer(data=data)
        self.assertFalse(serializer.is_valid())
        self.assertIn(
            ErrorDetail(string='Object with account_number=123456/0300 does not exist.', code='does_not_exist'),
            serializer.errors['bank_account'],
        )

    @freeze_time('2023-09-01T12:00:00')
    def test_account_not_compatible(self):
        create_payment_request(
            uuid='00000000-0000-0000-0000-000000000002',
            closed_at='2023-09-20T12:00:00',  # stored payment request is closed
        )
        BillingItemType.objects.create(
            handle='basic_rate', valid_from=datetime(2023, 1, 1), valid_to=None, vat_rate=Decimal('13.5')
        )

        data: Dict[str, Any] = {
            'uuid': '00000000-0000-0000-0000-000000000001',
            'app_id': 'auctions',
            'customer': 'CID-1',
            'due_at_displayed': '2023-10-01T12:00:00',
            'currency': 'USD',
            'variable_symbol': '0000009876',
            'bank_account': '123456/0300',
            'items': [
                {'name': 'Item 1', 'item_type': 'basic_rate', 'quantity': 1, 'price_per_unit': '11.0000000000',
                 'price_with_vat': True},
            ],
        }

        serializer = PaymentRequestSerializer(data=data)
        self.assertFalse(serializer.is_valid())
        self.assertIn(
            ErrorDetail(string='Selected currency is not compatible with the bank account', code='invalid'),
            serializer.errors['non_field_errors'],
        )

    @freeze_time('2023-09-01T12:00:00')
    def test_no_items(self):
        customer = get_customer()
        customer.save()
        account = get_account()
        account.save()

        data: Dict[str, Any] = {
            'uuid': '00000000-0000-0000-0000-000000000001',
            'app_id': 'auctions',
            'customer': 'CID-1',
            'due_at_displayed': '2023-10-01T12:00:00',
            'currency': 'CZK',
            'variable_symbol': '0000009876',
            'bank_account': '123456/0300',
            'items': [],
        }

        serializer = PaymentRequestSerializer(data=data)
        self.assertFalse(serializer.is_valid())
        self.assertIn(
            ErrorDetail(string='Ensure this field has at least 1 elements.', code='min_length'),
            serializer.errors['items']['non_field_errors'],
        )

    @freeze_time('2023-09-01T12:00:00')
    @override_settings(
        PAIN_PROCESSORS={'dummy': 'django_pain.tests.utils.DummyPaymentProcessor'},
        PAIN_CARD_PAYMENT_HANDLERS={
            'test': {'HANDLER': 'django_pain.tests.test_serializers.TestCardPaymentHandler', 'GATEWAY': 'default'},
        },
        PAIN_PAYMENT_REQUESTS={'auctions': {'CARD_HANDLER': 'test', 'PROCESSOR': 'dummy'}},
        PAIN_CSOB_CUSTOM_HANDLER={}
    )
    def test_create_card_payment_ok(self):
        create_payment_request(
            uuid='00000000-0000-0000-0000-000000000002',
            closed_at='2023-09-20T12:00:00',  # stored payment request is closed
        )
        BillingItemType.objects.create(
            handle='basic_rate', valid_from=datetime(2023, 1, 1), valid_to=None, vat_rate=Decimal('13.5')
        )

        data: Dict[str, Any] = {
            'uuid': '00000000-0000-0000-0000-000000000001',
            'app_id': 'auctions',
            'customer': 'CID-1',
            'due_at_displayed': '2023-10-01T12:00:00',
            'currency': 'CZK',
            'variable_symbol': '0000009876',
            'bank_account': '123456/0300',
            'items': [
                {'name': 'Item 1', 'item_type': 'basic_rate', 'quantity': 1, 'price_per_unit': '11.0000000000',
                 'price_with_vat': True},
                {'name': 'Item 2', 'item_type': 'basic_rate', 'quantity': 2, 'price_per_unit': '12.0000000000',
                 'price_with_vat': True},
                {'name': 'Item 3', 'item_type': 'basic_rate', 'quantity': 3, 'price_per_unit': '13.0000000000',
                 'price_with_vat': True},
            ],
            'create_card_payment': {
                'expiry': '2023-10-01T12:00:00',
                'redirect_url': 'https://callback.example.org/',
                'redirect_method': 'GET',
            }
        }

        serializer = PaymentRequestSerializer(data=data)
        self.assertTrue(serializer.is_valid())
        serializer.save()
        expected = get_payment_request_data(
            card_gateway_url='https://gateway.example.org/0000009876',
            card_payment=str(UUID(int=9876)),
            realized_payment=None,
        )
        self.assertEqual(serializer.data, expected)

    @freeze_time('2023-09-01T12:00:00')
    @override_settings(
        PAIN_PROCESSORS={'dummy': 'django_pain.tests.utils.DummyPaymentProcessor'},
        PAIN_CARD_PAYMENT_HANDLERS={
            'test': {'HANDLER': 'django_pain.tests.test_serializers.TestCardPaymentHandler', 'GATEWAY': 'default'},
        },
        PAIN_PAYMENT_REQUESTS={'auctions': {'CARD_HANDLER': 'test', 'PROCESSOR': 'dummy'}},
        PAIN_CSOB_CUSTOM_HANDLER={}
    )
    def test_create_card_payment_limit_exceeded(self):
        create_payment_request(
            uuid='00000000-0000-0000-0000-000000000002',
            closed_at='2023-09-20T12:00:00',  # stored payment request is closed
        )
        BillingItemType.objects.create(
            handle='basic_rate', valid_from=datetime(2023, 1, 1), valid_to=None, vat_rate=Decimal('13.5')
        )

        data: Dict[str, Any] = {
            'uuid': '00000000-0000-0000-0000-000000000001',
            'app_id': 'auctions',
            'customer': 'CID-1',
            'due_at_displayed': '2023-10-01T12:00:00',
            'currency': 'CZK',
            'variable_symbol': '0000009876',
            'bank_account': '123456/0300',
            'items': [
                {'name': 'Item 1', 'item_type': 'basic_rate', 'quantity': 1, 'price_per_unit': '11.0000000000',
                 'price_with_vat': True},
                {'name': 'Item 2', 'item_type': 'basic_rate', 'quantity': 2, 'price_per_unit': '12.0000000000',
                 'price_with_vat': True},
                {'name': 'Item 3', 'item_type': 'basic_rate', 'quantity': 3, 'price_per_unit': '13.0000000000',
                 'price_with_vat': True},
            ],
            'create_card_payment': {
                'expiry': '2023-10-01T12:00:00',
                'redirect_url': 'https://callback.example.org/',
                'redirect_method': 'GET',
            }
        }

        with patch.object(TestCardPaymentHandler, 'init_payment') as card_mock:
            card_mock.side_effect = PaymentHandlerLimitError

            serializer = PaymentRequestSerializer(data=data)
            self.assertTrue(serializer.is_valid())
            serializer.save()

            self.assertIsNone(serializer.instance.card_payment)  # card payment not created, no error reported

    # --- Making card payment cart ---

    def test_make_cart_one_item(self):
        items = [
            OrderedDict({'name': 'Item 1 with loooooong description', 'item_type': 'basic', 'quantity': 3,
                         'price_per_unit': Decimal('11.00'), 'price_total': Decimal('33.00')}),
        ]
        expected = [
            CartItem(name='Item 1 with loooooon', quantity=3, amount=33.0, description=''),
        ]

        serializer = PaymentRequestSerializer()
        self.assertEqual(serializer._make_cart(items), expected)

    def test_make_cart_two_items(self):
        items = [
            OrderedDict({'name': 'Item 1', 'item_type': 'basic', 'quantity': 3, 'price_per_unit': Decimal('11.00'),
                         'price_total': Decimal('33.00')}),
            OrderedDict({'name': 'Item 2', 'item_type': 'basic', 'quantity': 2, 'price_per_unit': Decimal('20.00'),
                         'price_total': Decimal('40.00')}),
        ]
        expected = [
            CartItem(name='Item 1', quantity=3, amount=33.0, description=''),
            CartItem(name='Item 2', quantity=2, amount=40.0, description=''),
        ]

        serializer = PaymentRequestSerializer()
        self.assertEqual(serializer._make_cart(items), expected)

    def test_make_cart_more_items(self):
        items = [
            OrderedDict({'name': 'Item 1', 'item_type': 'basic', 'quantity': 3, 'price_per_unit': Decimal('11.00'),
                         'price_total': Decimal('33.00')}),
            OrderedDict({'name': 'Item 2', 'item_type': 'basic', 'quantity': 1, 'price_per_unit': Decimal('20.00'),
                         'price_total': Decimal('20.00')}),
            OrderedDict({'name': 'Item 3', 'item_type': 'basic', 'quantity': 1, 'price_per_unit': Decimal('3.00'),
                         'price_total': Decimal('3.00')}),
            OrderedDict({'name': 'Item 4', 'item_type': 'basic', 'quantity': 1, 'price_per_unit': Decimal('4.00'),
                         'price_total': Decimal('4.00')}),
        ]
        expected = [
            CartItem(name='Item 1', quantity=3, amount=33.0, description=''),
            CartItem(name='+3 ...', quantity=1, amount=27.0, description=''),
        ]

        serializer = PaymentRequestSerializer()
        self.assertEqual(serializer._make_cart(items), expected)


class TestPaymentRequestContextSerializer(TestCase):
    """`PaymentRequestContextSerializer` tests."""

    def test_serialization(self):
        # Prepare data and save all objects to database
        pk = create_payment_request().pk

        # Get object from database and serialize it
        serializer = PaymentRequestContextSerializer(PaymentRequest.objects.get(pk=pk))
        self.assertEqual(serializer.data, get_payment_request_context_data())

    def test_iban_ok(self):
        data = [
            ('276463778/0300', 'CZ3703000000000276463778'),
            ('625/5500', 'CZ3155000000000000000625'),
            ('000000-0000000625/5500', 'CZ3155000000000000000625'),
            ('268518759/ 0300', 'CZ5003000000000268518759'),
            ('268518759 /0300', 'CZ5003000000000268518759'),
            ('268518759 / 0300', 'CZ5003000000000268518759'),
            ('288234710/0300', 'CZ3803000000000288234710'),
            ('19-288234710/0300', 'CZ2403000000190288234710'),
            ('20001- 288234710/0300', 'CZ2403000200010288234710'),
            ('20001 -288234710/0300', 'CZ2403000200010288234710'),
            ('20001 - 288234710/0300', 'CZ2403000200010288234710'),
        ]
        for account, iban in data:
            with self.subTest(account=account):
                serializer = PaymentRequestContextSerializer()
                self.assertEqual(serializer._account_to_iban(account), iban)

    def test_iban_exception(self):
        data = [
            '',
            'A10/0100',
            '123//0100',
            '1/0100',
            '98765432109/0100',
            '9*123/0100',
            '1-123/0100',
            '1234567-123/0100',
            '123/105',
        ]
        for account in data:
            with self.subTest(account=account):
                serializer = PaymentRequestContextSerializer()
                self.assertRaises(ValueError, serializer._account_to_iban, account)
