#
# Copyright (C) 2018-2023  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Test utils."""
from datetime import date, datetime, timezone
from decimal import ROUND_DOWN, Decimal
from typing import Dict, List, Tuple

from django.test import SimpleTestCase, override_settings

from django_pain.constants import PaymentProcessingError, ProcessPaymentResultType
from django_pain.models.bank import BankAccount
from django_pain.utils import (ProcessPaymentResult, full_class_name, local_date, parse_date_safe, parse_datetime_safe,
                               round_money_amount)


class FullClassNameTest(SimpleTestCase):

    def test_str(self):
        """Test full_class_name."""
        self.assertEqual(full_class_name(BankAccount), 'django_pain.models.bank.BankAccount')


class ParseDateSafeTest(SimpleTestCase):

    def test_parse_date_safe(self):
        self.assertEqual(parse_date_safe('2017-01-31'), date(2017, 1, 31))

    def test_parse_date_safe_fails_on_invalid(self):
        with self.assertRaises(ValueError):
            parse_date_safe('2017-01-32')
        with self.assertRaises(ValueError):
            parse_date_safe('not a date')


class ParseDatetimeSafeTest(SimpleTestCase):

    def test_parse_datetime_safe(self):
        self.assertEqual(parse_datetime_safe('2017-01-31 00:00'), datetime(2017, 1, 31, 0, 0))

    def test_parse_datetime_safe_fails_on_invalid(self):
        with self.assertRaises(ValueError):
            parse_datetime_safe('2017-01-32 00:00')
        with self.assertRaises(ValueError):
            parse_datetime_safe('not a date')


class RoundMoneyAmountTest(SimpleTestCase):

    def test_rounding(self):
        data: List[Tuple[Dict, str]] = [
            ({'amount': Decimal('1.1')}, '1.10'),
            ({'amount': Decimal('1.124')}, '1.12'),
            ({'amount': Decimal('1.125')}, '1.13'),
            ({'amount': Decimal('1.126')}, '1.13'),
            ({'amount': Decimal('1.124'), 'places': 1}, '1.1'),
            ({'amount': Decimal('1.119'), 'rounding': ROUND_DOWN}, '1.11'),
        ]
        for param, expected in data:
            with self.subTest(param=param):
                self.assertEqual(str(round_money_amount(**param)), expected)


class LocalDateTest(SimpleTestCase):

    @override_settings(TIME_ZONE='Europe/Prague')
    def test_localdate(self):
        data = [
            # naive datetimes
            (datetime(2023, 10, 1, 0, 0, 0), date(2023, 10, 1)),
            (datetime(2023, 10, 1, 23, 59, 59), date(2023, 10, 1)),
            # aware datetimes
            (datetime(2023, 10, 1, 0, 0, 0, tzinfo=timezone.utc), date(2023, 10, 1)),
            (datetime(2023, 10, 1, 23, 59, 59, tzinfo=timezone.utc), date(2023, 10, 2)),  # timezone makes difference!
        ]
        for dt, expected in data:
            with self.subTest(datetime=dt):
                self.assertEqual(local_date(dt), expected)


class TestProcessPaymentResult(SimpleTestCase):
    """Test ProcessPaymentResult."""

    def test_process_payment_result_init(self):
        """Test ProcessPaymentResult __init__."""
        self.assertEqual(ProcessPaymentResult(ProcessPaymentResultType.PROCESSED).result,
                         ProcessPaymentResultType.PROCESSED)
        self.assertEqual(ProcessPaymentResult(ProcessPaymentResultType.TO_REFUND).result,
                         ProcessPaymentResultType.TO_REFUND)
        self.assertEqual(ProcessPaymentResult(ProcessPaymentResultType.NOT_PROCESSED).result,
                         ProcessPaymentResultType.NOT_PROCESSED)
        self.assertEqual(ProcessPaymentResult(ProcessPaymentResultType.PROCESSED).error, None)
        self.assertEqual(
            ProcessPaymentResult(ProcessPaymentResultType.PROCESSED, PaymentProcessingError.DUPLICITY).error,
            PaymentProcessingError.DUPLICITY
        )

    def test_process_payment_result_eq(self):
        """Test ProcessPaymentResult __eq__."""
        self.assertEqual(ProcessPaymentResult(ProcessPaymentResultType.PROCESSED),
                         ProcessPaymentResult(ProcessPaymentResultType.PROCESSED))
        self.assertNotEqual(ProcessPaymentResult(ProcessPaymentResultType.PROCESSED),
                            ProcessPaymentResult(ProcessPaymentResultType.NOT_PROCESSED))
        self.assertNotEqual(ProcessPaymentResult(ProcessPaymentResultType.NOT_PROCESSED), 0)
        self.assertNotEqual(ProcessPaymentResult(ProcessPaymentResultType.NOT_PROCESSED),
                            ProcessPaymentResult(ProcessPaymentResultType.NOT_PROCESSED,
                                                 PaymentProcessingError.DUPLICITY))

    def test_unprocessed(self):
        """Test detecting uprocessed state."""
        data = [
            (ProcessPaymentResultType.PROCESSED, None, False),
            (ProcessPaymentResultType.TO_REFUND, None, False),
            (ProcessPaymentResultType.NOT_PROCESSED, PaymentProcessingError.DUPLICITY, False),
            (ProcessPaymentResultType.NOT_PROCESSED, None, True),
        ]
        for result, error, expected in data:
            with self.subTest(result=result, error=error):
                instance = ProcessPaymentResult(result=result, error=error)
                self.assertEqual(instance.unprocessed, expected)
