#
# Copyright (C) 2018-2023  CZ.NIC, z. s. p. o.
#
# This file is part of FRED.
#
# FRED is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FRED is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FRED.  If not, see <https://www.gnu.org/licenses/>.

"""Various utils."""
from datetime import date, datetime
from decimal import ROUND_HALF_UP, Decimal
from typing import Optional

from django.utils import timezone
from django.utils.dateparse import parse_date, parse_datetime
from djmoney.settings import DECIMAL_PLACES

from django_pain.constants import PaymentProcessingError, ProcessPaymentResultType


def full_class_name(cls):
    """Return full class name includeing the module path."""
    return "{}.{}".format(cls.__module__, cls.__qualname__)


def parse_date_safe(value: str) -> date:
    """Parse date using Django utils, but raise an exception when unsuccessful."""
    result = parse_date(value)
    if result is None:
        raise ValueError('Could not parse date.')
    return result


def parse_datetime_safe(value: str) -> datetime:
    """Parse date_time using Django utils, but raise an exception when unsuccessful."""
    result = parse_datetime(value)
    if result is None:
        raise ValueError('Could not parse date_time.')
    return result


def round_money_amount(amount: Decimal, places: int = DECIMAL_PLACES, rounding: str = ROUND_HALF_UP) -> Decimal:
    """
    Round money amount to given places and using given rounding method.

    Provided defaults should be OK for Czechia, at least (it may be worth noting that currency-specific decimal
    places are not implemented even in django-money which uses hard-coded value for money formatting).
    """
    q = Decimal(10) ** -places
    return amount.quantize(q, rounding=rounding)


def local_date(value: datetime) -> date:
    """Get date from timestamp considering local time zone."""
    return value.astimezone(timezone.get_current_timezone()).date()


class ProcessPaymentResult(object):
    """Result of payment processing."""

    def __init__(self, result: ProcessPaymentResultType, error: Optional[PaymentProcessingError] = None) -> None:
        """
        Initialize the result.

        Args:
            result: PROCESSED if payment was successfully processed, TO_REFUND if bound for refunding,
                    NOT_PROCESSED otherwise.
            error: Optional payment processing error code.
        """
        self.result = result
        self.error = error

    def __eq__(self, other) -> bool:
        """Compare processing results by their value."""
        if isinstance(other, ProcessPaymentResult):
            return self.result == other.result and self.error == other.error
        return False

    @property
    def unprocessed(self) -> bool:
        """
        Return whether the payment is left unprocessed.

        The state with `NOT_PROCESSED` result and no error represents case in which processor didn't handle
        the payment (and left it for other processors or handling at a later time).
        """
        return self.result == ProcessPaymentResultType.NOT_PROCESSED and self.error is None
